@@ -52,10 +52,13 @@ THE SOFTWARE.
 #include "resend.h"
 #include "configuration.h"
 #include "local.h"
+#include "rule.h"
+#include "version.h"
 
 struct timeval now;
 
 unsigned char myid[8];
+int have_id = 0;
 int debug = 0;
 
 int link_detect = 0;
@@ -66,6 +69,7 @@ int default_wired_hello_interval = -1;
 int resend_delay = -1;
 int random_id = 0;
 int do_daemonise = 0;
+int skip_kernel_setup = 0;
 const char *logfile = NULL,
     *pidfile = "/var/run/babeld.pid",
     *state_file = "/var/lib/babel-state";
@@ -92,11 +96,53 @@ struct timeval check_neighbours_timeout, check_interfaces_timeout;
 static volatile sig_atomic_t exiting = 0, dumping = 0, reopening = 0;
 
 static int accept_local_connections(fd_set *readfds);
-static int kernel_routes_callback(int changed, void *closure);
 static void init_signals(void);
 static void dump_tables(FILE *out);
 static int reopen_logfile(void);
 
+static int
+kernel_route_notify(struct kernel_route *route, void *closure)
+{
+    kernel_routes_changed = 1;
+    return -1;
+}
+
+static int
+kernel_addr_notify(struct kernel_addr *addr, void *closure)
+{
+    kernel_addr_changed = 1;
+    return -1;
+}
+
+static int
+kernel_link_notify(struct kernel_link *link, void *closure)
+{
+    struct interface *ifp;
+    FOR_ALL_INTERFACES(ifp) {
+        if(strcmp(ifp->name, link->ifname) == 0) {
+            kernel_link_changed = 1;
+            return -1;
+        }
+    }
+    return 0;
+}
+
+static int
+kernel_rule_notify(struct kernel_rule *rule, void *closure)
+{
+    int i;
+    if(martian_prefix(rule->src, rule->src_plen))
+        return 0;
+
+    i = rule->priority - src_table_prio;
+
+    if(i < 0 || SRC_TABLE_NUM <= i)
+        return 0;
+
+    kernel_rules_changed = 1;
+    return -1;
+}
+
 int
 main(int argc, char **argv)
 {
@@ -126,7 +172,8 @@ main(int argc, char **argv)
     has_ipv6_subtrees = kernel_has_ipv6_subtrees();
 
     while(1) {
-        opt = getopt(argc, argv, "m:p:h:H:i:k:A:sruS:d:g:lwz:M:t:T:c:C:DL:I:");
+        opt = getopt(argc, argv,
+                     "m:p:h:H:i:k:A:sruS:d:g:lwz:M:t:T:c:C:DL:I:V");
         if(opt < 0)
             break;
 
@@ -261,6 +308,10 @@ main(int argc, char **argv)
         case 'I':
             pidfile = optarg;
             break;
+        case 'V':
+            fprintf(stderr, "%s\n", BABELD_VERSION);
+            exit(0);
+            break;
         default:
             goto usage;
         }
@@ -395,50 +446,40 @@ main(int argc, char **argv)
         goto fail;
     }
 
-    if(random_id)
-        goto random_id;
+    if(!have_id && !random_id) {
+        /* We use all available interfaces here, since this increases the
+           chances of getting a stable router-id in case the set of Babel
+           interfaces changes. */
 
-    FOR_ALL_INTERFACES(ifp) {
-        /* ifp->ifindex is not necessarily valid at this point */
-        int ifindex = if_nametoindex(ifp->name);
-        if(ifindex > 0) {
+        for(i = 1; i < 256; i++) {
+            char buf[IF_NAMESIZE], *ifname;
             unsigned char eui[8];
-            rc = if_eui64(ifp->name, ifindex, eui);
+            ifname = if_indextoname(i, buf);
+            if(ifname == NULL)
+                continue;
+            rc = if_eui64(ifname, i, eui);
             if(rc < 0)
                 continue;
             memcpy(myid, eui, 8);
-            goto have_id;
+            have_id = 1;
+            break;
         }
     }
 
-    /* We failed to get a global EUI64 from the interfaces we were given.
-       Let's try to find an interface with a MAC address. */
-    for(i = 1; i < 256; i++) {
-        char buf[IF_NAMESIZE], *ifname;
-        unsigned char eui[8];
-        ifname = if_indextoname(i, buf);
-        if(ifname == NULL)
-            continue;
-        rc = if_eui64(ifname, i, eui);
-        if(rc < 0)
-            continue;
-        memcpy(myid, eui, 8);
-        goto have_id;
+    if(!have_id) {
+        if(!random_id)
+            fprintf(stderr,
+                    "Warning: couldn't find router id -- "
+                    "using random value.\n");
+        rc = read_random_bytes(myid, 8);
+        if(rc < 0) {
+            perror("read(random)");
+            goto fail;
+        }
+        /* Clear group and global bits */
+        myid[0] &= ~3;
     }
 
-    fprintf(stderr,
-            "Warning: couldn't find router id -- using random value.\n");
-
- random_id:
-    rc = read_random_bytes(myid, 8);
-    if(rc < 0) {
-        perror("read(random)");
-        goto fail;
-    }
-    /* Clear group and global bits */
-    myid[0] &= ~3;
-
- have_id:
     myseqno = (random() & 0xFFFF);
 
     fd = open(state_file, O_RDONLY);
@@ -453,30 +494,15 @@ main(int argc, char **argv)
     }
     if(fd >= 0) {
         char buf[100];
-        char buf2[100];
         int s;
-        long t;
         rc = read(fd, buf, 99);
         if(rc < 0) {
             perror("read(babel-state)");
         } else {
             buf[rc] = '\0';
-            rc = sscanf(buf, "%99s %d %ld\n", buf2, &s, &t);
-            if(rc == 3 && s >= 0 && s <= 0xFFFF) {
-                unsigned char sid[8];
-                rc = parse_eui64(buf2, sid);
-                if(rc < 0) {
-                    fprintf(stderr, "Couldn't parse babel-state.\n");
-                } else {
-                    struct timeval realnow;
-                    debugf("Got %s %d %ld from babel-state.\n",
-                           format_eui64(sid), s, t);
-                    gettimeofday(&realnow, NULL);
-                    if(memcmp(sid, myid, 8) == 0)
-                        myseqno = seqno_plus(s, 1);
-                    else if(!random_id)
-                        fprintf(stderr, "ID mismatch in babel-state.\n");
-                }
+            rc = sscanf(buf, "%d\n", &s);
+            if(rc == 1 && s >= 0 && s <= 0xFFFF) {
+                myseqno = seqno_plus(s, 1);
             } else {
                 fprintf(stderr, "Couldn't parse babel-state.\n");
             }
@@ -513,6 +539,9 @@ main(int argc, char **argv)
     rc = check_xroutes(0);
     if(rc < 0)
         fprintf(stderr, "Warning: couldn't check exported routes.\n");
+    rc = check_rules();
+    if(rc < 0)
+        fprintf(stderr, "Warning: couldn't check rules.\n");
 
     kernel_routes_changed = 0;
     kernel_rules_changed = 0;
@@ -610,8 +639,14 @@ main(int argc, char **argv)
         if(exiting)
             break;
 
-        if(kernel_socket >= 0 && FD_ISSET(kernel_socket, &readfds))
-            kernel_callback(kernel_routes_callback, NULL);
+        if(kernel_socket >= 0 && FD_ISSET(kernel_socket, &readfds)) {
+            struct kernel_filter filter = {0};
+            filter.route = kernel_route_notify;
+            filter.addr = kernel_addr_notify;
+            filter.link = kernel_link_notify;
+            filter.rule = kernel_rule_notify;
+            kernel_callback(&filter);
+        }
 
         if(FD_ISSET(protocol_socket, &readfds)) {
             rc = babel_recv(protocol_socket,
@@ -681,6 +716,9 @@ main(int argc, char **argv)
             rc = check_xroutes(1);
             if(rc < 0)
                 fprintf(stderr, "Warning: couldn't check exported routes.\n");
+            rc = check_rules();
+            if(rc < 0)
+                fprintf(stderr, "Warning: couldn't check rules.\n");
             kernel_routes_changed = kernel_rules_changed =
                 kernel_addr_changed = 0;
             if(kernel_socket >= 0)
@@ -778,6 +816,7 @@ main(int argc, char **argv)
         gettime(&now);
         interface_up(ifp, 0);
     }
+    release_tables();
     kernel_setup_socket(0);
     kernel_setup(0);
 
@@ -786,13 +825,9 @@ main(int argc, char **argv)
         perror("creat(babel-state)");
         unlink(state_file);
     } else {
-        struct timeval realnow;
-        char buf[100];
-        gettimeofday(&realnow, NULL);
-        rc = snprintf(buf, 100, "%s %d %ld\n",
-                      format_eui64(myid), (int)myseqno,
-                      (long)realnow.tv_sec);
-        if(rc < 0 || rc >= 100) {
+        char buf[10];
+        rc = snprintf(buf, 10, "%d\n", (int)myseqno);
+        if(rc < 0 || rc >= 10) {
             fprintf(stderr, "write(babel-state): overflow.\n");
             unlink(state_file);
         } else {
@@ -812,19 +847,20 @@ main(int argc, char **argv)
 
  usage:
     fprintf(stderr,
-            "Syntax: %s "
-            "[-m multicast_address] [-p port] [-S state-file]\n"
-            "                "
+            "%s\n"
+            "Syntax: babeld "
+            "[-V] [-m multicast_address] [-p port] [-S state-file]\n"
+            "               "
             "[-h hello] [-H wired_hello] [-z kind[,factor]]\n"
-            "                "
+            "               "
             "[-k metric] [-A metric] [-s] [-l] [-w] [-r] [-u] [-g port]\n"
-            "                "
+            "               "
             "[-t table] [-T table] [-c file] [-C statement]\n"
-            "                "
+            "               "
             "[-d level] [-D] [-L logfile] [-I pidfile]\n"
-            "                "
+            "               "
             "[id] interface...\n",
-            argv[0]);
+            BABELD_VERSION);
     exit(1);
 
  fail:
@@ -1093,7 +1129,7 @@ dump_tables(FILE *out)
         xroute_stream_done(xroutes);
     }
 
-    routes = route_stream(0);
+    routes = route_stream(ROUTE_ALL);
     if(routes) {
         while(1) {
             struct babel_route *route = route_stream_next(routes);
@@ -1134,17 +1170,3 @@ reopen_logfile()
 
     return 1;
 }
-
-static int
-kernel_routes_callback(int changed, void *closure)
-{
-    if(changed & CHANGE_LINK)
-        kernel_link_changed = 1;
-    if(changed & CHANGE_ADDR)
-        kernel_addr_changed = 1;
-    if(changed & CHANGE_ROUTE)
-        kernel_routes_changed = 1;
-    if(changed & CHANGE_RULE)
-        kernel_rules_changed = 1;
-    return 1;
-}
