@@ -86,6 +86,7 @@ extern time_t reboot_time;
 extern int default_wireless_hello_interval, default_wired_hello_interval;
 extern int resend_delay;
 extern int random_id;
+extern int skip_kernel_setup;
 extern int do_daemonise;
 extern const char *logfile, *pidfile, *state_file;
 extern int link_detect;
@@ -93,6 +94,7 @@ extern int all_wireless;
 extern int has_ipv6_subtrees;
 
 extern unsigned char myid[8];
+extern int have_id;
 
 extern const unsigned char zeroes[16], ones[16];
 
